@@ -23,10 +23,13 @@
 import random
 import urllib2
 import httplib
+import socket
 import datetime
 import re
 import xml.sax
 import xml.sax.saxutils
+import xml.dom
+import xml.dom.pulldom
 import StringIO
 import HTMLParser
 import htmlentitydefs
@@ -37,95 +40,114 @@ try:
 except NameError:
     from sets import Set as set
 
-import phpserialize
+try:
+    import phpserialize
+except ImportError:
+    phpserialize = None
 
 from linkexchange.clients.base import BaseClient, SimpleFileTestServer
 from linkexchange.clients.base import ClientError, \
         ClientNetworkError, ClientDataError, ClientDataAccessError
 from linkexchange.utils import is_plugin_specifier, load_plugin
-from linkexchange.utils import urlopen_with_timeout, get_default_user_agent
+from linkexchange.utils import urlopen_with_timeout, default_user_agent
+from linkexchange.utils import normalize_uri
 
 log = logging.getLogger('linkexchange.clients.sape')
 
-class SapeTestServer(SimpleFileTestServer):
-    filename = 'sape_test_server_data.txt'
-    data = {
-        '/': [
-            '<a href="url1">link1</a>',
-            '<a href="url2">link2</a>'],
-        '/path/1': [
-            '<a href="url1">link1</a>',
-            '<a href="url2">link2</a>',
-            '<a href="url3">link3</a>',
-            '<a href="url4">link4</a>'],
-        '/path/2': [
-            'Plain text and <a href="url">link text</a>'],
-        '__sape_new_url__': '<!--12345-->'}
+class SapeLikeTestServer(SimpleFileTestServer):
+    data = None
     extra_data = None
+    use_xml = True
 
-    def __init__(self, filename = None, data = None, extra_data = None):
+    def __init__(self, filename = None, data = None, extra_data = None,
+            use_xml = None):
         if data is not None:
             self.data = data
         if extra_data is not None:
             self.extra_data = extra_data
         if self.extra_data:
             self.data.update(self.extra_data)
-        raw_data = phpserialize.dumps(self.data)
-        super(SapeTestServer, self).__init__(filename = filename,
+        if use_xml is not None:
+            self.use_xml = use_xml
+        if self.use_xml:
+            raw_data = self.format_data_xml(self.data)
+        else:
+            raw_data = self.format_data(self.data)
+        super(SapeLikeTestServer, self).__init__(filename = filename,
                 raw_data = raw_data)
 
-class SapeClient(BaseClient):
-    """
-    Sape.ru client.
+    def format_data(self, data):
+        return phpserialize.dumps(data)
 
-    >>> from linkexchange.clients.base import PageRequest
-    >>> test_server = SapeTestServer()
-    >>> cl = SapeClient(user = 'user123456789', db_driver = ('mem',),
-    ...                 server_list = [test_server.url])
-    >>> lx = cl.get_raw_links(PageRequest(url = 'http://example.com/'))
-    >>> lx[0]
-    u'<a href="url1">link1</a>'
-    >>> lx[1]
-    u'<a href="url2">link2</a>'
-    >>> cl = SapeClient(user = 'user123456789', db_driver = ('mem',),
-    ...                 server_list = ['file:///does_not_exists'])
-    >>> cl.get_raw_links(PageRequest(url = 'http://example.com/'))
-    []
-    >>> cl.get_raw_links(PageRequest(url = 'http://example.com/'))
-    []
+    def format_data_xml(self, data):
+        return ''
+
+class SapeLikeClient(BaseClient):
     """
+    Base class for Sape-like clients.
+    """
+    db_lifetime = datetime.timedelta(seconds = 3600)
+    db_reloadtime = datetime.timedelta(seconds = 600)
+    socket_timeout = 6
+    force_show_code = True
+    server_list = []
+    xml_server_list = []
+    use_xml = True
+    user_agent = default_user_agent
+
     def __init__(self, user, db_driver, **kw):
         """
-        @param user: user ID of Sape account
+        SapeLikeClient constructor.
+        
+        The user is hash code string that assigned to user on link exchange
+        service.
+
+        The db_driver argument is multihash database driver instance or plugin
+        specifier. In second case plugin specifier is used to create new
+        instance. The database driver instance is used to store links database.
+
+        The db_lifetime keyword argument specifies database lifetime. Database
+        older than this time interval will be updated be calling refresh_db().
+        Value can be datetime.timedelta object or number of seconds as integer
+        or float value. None value disables database refreshing, but refreshing
+        will occur if database is not exist even if db_lifetime is None.
+
+        The db_reloadtime specifies the time interval during which the database
+        refreshing is not happens if the previous attempt to refreshing was
+        failed. This prevents remote server overloading when problem on client
+        side. Value can be datetime.timedelta object or number of seconds as
+        integer or float value. None value means that db_lifetime value should
+        be used.
+
+        The socket_timeout is socket timeout for remote connections in seconds.
+
+        @param user: user hash code string on link exchange service
         @param db_driver: multihash database driver instance or plugin specifier
         @keyword db_lifetime: DB lifetime as datetime.timedelta object or
-                              number of seconds as numeric value
-        @keyword db_reloadtime: if an error in DB refreshing occurred don't
-                                refresh DB within this time (datetime.timedelta
-                                or number of seconds)
+                              number of seconds as numeric value or None
+        @keyword db_reloadtime: DB reload time as datetime.timedelta object or
+                              number of seconds as numeric value or None
         @keyword socket_timeout: socket timeout in seconds
         @keyword force_show_code: if True force to show check code
-        @keyword server_list: list of servers URLs
+        @keyword server_list: list of servers URLs that returns PHP data
+        @keyword xml_server_list: list of servers URLs that returns XML data
+        @keyword use_xml: fetch database in XML format from servers in
+                          xml_server_list
         @keyword user_agent: user agent string
         """
         self.user = user
         if is_plugin_specifier(db_driver):
             db_driver = load_plugin('linkexchange.multihash_drivers', db_driver)
         self.db_driver = db_driver
-        self.db_lifetime = kw.get('db_lifetime',
-                datetime.timedelta(seconds = 3600))
-        if type(self.db_lifetime) in (int, long, float):
-            self.db_lifetime = datetime.timedelta(seconds = self.db_lifetime)
-        self.db_reloadtime = kw.get('db_reloadtime',
-                datetime.timedelta(seconds = 600))
-        if type(self.db_reloadtime) in (int, long, float):
-            self.db_reloadtime = datetime.timedelta(seconds = self.db_reloadtime)
-        self.socket_timeout = kw.get('socket_timeout', 6)
-        self.force_show_code = kw.get('force_show_code', True)
-        self.server_list = kw.get('server_list', [
-            'http://dispenser-01.sape.ru/code.php?user=%(user)s&host=%(host)s&charset=utf-8',
-            'http://dispenser-02.sape.ru/code.php?user=%(user)s&host=%(host)s&charset=utf-8'])
-        self.user_agent = kw.get('user_agent', get_default_user_agent())
+        for param in ('db_lifetime', 'db_reloadtime', 'socket_timeout',
+                'force_show_code', 'server_list', 'xml_server_list',
+                'use_xml', 'user_agent'):
+            if param in kw:
+                setattr(self, param, kw[param])
+        for param in ('db_lifetime', 'db_reloadtime'):
+            value = getattr(self, param)
+            if type(value) in (int, long, float):
+                setattr(self, param, datetime.timedelta(seconds = value))
         log.debug("New %s instance:\n%s",
                 self.__class__.__name__,
                 '\n'.join(["    %s: %s" % (x, repr(getattr(self, x)))
@@ -137,15 +159,17 @@ class SapeClient(BaseClient):
                         'socket_timeout',
                         'force_show_code',
                         'server_list',
+                        'xml_server_list',
+                        'use_xml',
                         'user_agent')]))
 
-    def _normalize_host(self, request):
-        host = request.host
+    def normalize_host(self, request):
+        host = request.host.lower()
         if host.startswith('www.'):
             host = host[len('www.'):]
         return host
 
-    def _load_data(self, request):
+    def load_data(self, request):
         def save_error(host, data, error):
             new_data = {}
             new_data.update(data)
@@ -153,24 +177,25 @@ class SapeClient(BaseClient):
             new_data['__error_value__'] = error
             return self.db_driver.save(host, new_data, blocking = False)
 
-        host = self._normalize_host(request)
+        host = self.normalize_host(request)
         data = None
         force_refresh = False
         try:
             data = self.db_driver.load(host)
         except KeyError:
             log.debug("No existing database found, creating new one")
-        else:
+        if data is not None and self.db_lifetime is not None:
+            reloadtime = self.db_reloadtime
+            if reloadtime is None:
+                reloadtime = self.db_lifetime
             try:
-                refresh_after = data['__error_time__'] + self.db_reloadtime
+                refresh_after = data['__error_time__'] + reloadtime
             except KeyError:
-                refresh_after = self.db_driver.get_mtime(host) + self.db_lifetime
+                refresh_after = (self.db_driver.get_mtime(host) +
+                        self.db_lifetime)
             if refresh_after <= datetime.datetime.now():
                 log.debug("The database too old, refreshing")
                 force_refresh = True
-            elif type(data.get('__sape_new_url__', None)) == str:
-                # database from old version
-                data = None
         if data is None:
             try:
                 self.refresh_db(request)
@@ -188,41 +213,56 @@ class SapeClient(BaseClient):
             data = self.db_driver.load(host)
         return data
 
-    def _get_links(self, data, request):
+    def get_links(self, data, request):
         try:
             return data[str(request.uri)]
         except KeyError:
-            if (request.cookies.get('sape_cookie') == self.user or
-                    self.force_show_code):
-                try:
-                    return [data['__sape_new_url__']]
-                except KeyError:
-                    pass
+            return self.get_links_new_page(data, request)
+
+    def get_links_new_page(self, data, request):
         return []
 
+    def get_delimiter(self, data, request):
+        return ''
+
+    def is_bot(self, data, request):
+        return False
+
+    def transform_code(self, data, request, code):
+        return code
+
     def get_raw_links(self, request):
         log.debug("Getting raw links for: %s", request.url())
-        data = self._load_data(request)
-        links = self._get_links(data, request)
-        if request.cookies.get('sape_cookie') == self.user:
-            links = ['<sape_noindex>%s</sape_noindex>' % x for x in links]
-        return links
+        data = self.load_data(request)
+        links = self.get_links(data, request)
+        return [self.transform_code(data, request, code) for code in links]
 
     def get_html_links(self, request):
         log.debug("Getting HTML links for: %s", request.url())
-        data = self._load_data(request)
-        links = self._get_links(data, request)
-        delim = data.get('__sape_delimiter__', '')
+        data = self.load_data(request)
+        links = self.get_links(data, request)
+        delim = self.get_delimiter(data, request)
         html = delim.join(links)
-        if request.cookies.get('sape_cookie') == self.user:
-            html = '<sape_noindex>%s</sape_noindex>' % html
-        return html
+        return self.transform_code(data, request, html)
 
-    def _do_load_from_server(self, url):
-        def str2uni(value):
-            if type(value) == str:
-                value = unicode(value, 'utf-8')
-            return value
+    def parse_link(self, link):
+        if type(link) == str:
+            link = unicode(link, 'utf-8')
+        return link
+
+    def parse_data(self, data):
+        for key, value in data.items():
+            if key.startswith('/'):
+                if type(value) == dict:
+                    value = value.values()
+                yield (normalize_uri(key), map(self.parse_link, value))
+            else:
+                if type(value) == str:
+                    value = unicode(value, 'utf-8')
+                yield (key, value)
+
+    def load_from_server(self, server, host):
+        url = server % dict(user = self.user, host = host)
         req = urllib2.Request(url)
         req.add_header('User-Agent', self.user_agent)
         try:
@@ -230,29 +270,49 @@ class SapeClient(BaseClient):
             f = urlopen_with_timeout(req, self.socket_timeout)
             raw_data = f.read()
             f.close()
-        except (urllib2.URLError, httplib.HTTPException, OSError), e:
-            log.error("Network error: %s", str(e))
+        except (urllib2.URLError, httplib.HTTPException, OSError,
+                socket.error), e:
+            log.error("Network error: %s: %s", str(e), url)
             raise ClientNetworkError('Network error: %s' % str(e))
         if raw_data.startswith('FATAL ERROR:'):
-            log.error("Server error: %s", raw_data)
+            log.error("Server error: %s: %s", raw_data, url)
             raise ClientError(raw_data)
         try:
             data = phpserialize.loads(raw_data)
         except ValueError, e:
-            log.error("Could not deserialize response from server: %s", str(e))
+            log.error("Could not deserialize response from server: %s: %s", str(e), url)
             raise ClientDataError('Could not deserialize response '
                     'from server: %s' % str(e))
-        for key, value in data.items():
-            if key.startswith('/'):
-                if type(value) == dict:
-                    value = value.values()
-                yield (key, map(str2uni, value))
-            else:
-                yield (key, str2uni(value))
+        return self.parse_data(data)
+
+    def parse_data_xml(self, data):
+        return []
+
+    def load_from_server_xml(self, server, host):
+        url = server % dict(user = self.user, host = host)
+        req = urllib2.Request(url)
+        req.add_header('User-Agent', self.user_agent)
+        try:
+            log.debug("Fetching: %s", url)
+            f = urlopen_with_timeout(req, self.socket_timeout)
+            events = xml.dom.pulldom.parse(f)
+            for x in self.parse_data_xml(events):
+                yield x
+            f.close()
+        except (urllib2.URLError, httplib.HTTPException, OSError,
+                socket.error), e:
+            log.error("Network error: %s: %s", str(e), url)
+            raise ClientNetworkError('Network error: %s' % str(e))
+        except xml.sax.SAXParseException, e:
+            log.error("Could not parse XML data: %s: %s", str(e), url)
+            raise ClientDataError('Could not parse XML data: %s' % str(e))
 
     def refresh_db(self, request):
-        host = self._normalize_host(request)
-        server_list = self.server_list[:]
+        host = self.normalize_host(request)
+        if self.use_xml:
+            server_list = self.xml_server_list[:]
+        else:
+            server_list = self.server_list[:]
         random.shuffle(server_list)
         server_list = iter(server_list)
         data = None
@@ -260,9 +320,11 @@ class SapeClient(BaseClient):
 
         while data is None:
             try:
-                server = server_list.next() % dict(
-                        user = self.user, host = host)
-                data = self._do_load_from_server(server)
+                server = server_list.next()
+                if self.use_xml:
+                    data = self.load_from_server_xml(server, host)
+                else:
+                    data = self.load_from_server(server, host)
             except StopIteration:
                 raise error
             except ClientError, e:
@@ -273,6 +335,131 @@ class SapeClient(BaseClient):
             raise ClientDataAccessError(
                     "Other process/tread is currently writes to the database")
 
+class SapeTestServer(SapeLikeTestServer):
+    filename = 'sape_test_server_data.txt'
+    data = {
+        '/': [
+            '<a href="url1">link1</a>',
+            '<a href="url2">link2</a>'],
+        '/path/1': [
+            '<a href="url1">link1</a>',
+            '<a href="url2">link2</a>',
+            '<a href="url3">link3</a>',
+            '<a href="url4">link4</a>'],
+        '/path/2': [
+            'Plain text and <a href="url">link text</a>'],
+        '__sape_new_url__': '<!--12345-->',
+        '__sape_delimiter__': '. ',
+        }
+
+    def format_data_xml(self, data):
+        def make_page(uri, links):
+            return '<page uri="%s">%s</page>' % (uri,
+                    ''.join(['<link><![CDATA[%s]]></link>' % s
+                        for s in links]))
+
+        pages = '\n'.join([make_page(uri, links)
+            for uri, links in data.items() if uri.startswith('/')])
+
+        lines = [
+                '<?xml version="1.0" encoding="UTF-8"?>',
+                '<sape delimiter="%s">' % data.get('__sape_delimiter__', ''),
+                pages,
+                '<page uri="*"><![CDATA[%s]]></page>' % data.get(
+                    '__sape_new_url__', ''),
+                '</sape>']
+        return '\n'.join(lines)
+
+class SapeClient(SapeLikeClient):
+    """
+    Sape.ru client.
+
+    >>> from linkexchange.clients.base import PageRequest
+    >>> test_server = SapeTestServer()
+    >>> cl = SapeClient(user = 'user123456789', db_driver = ('mem',),
+    ...                 use_xml = True, xml_server_list = [test_server.url])
+    >>> lx = cl.get_raw_links(PageRequest(url = 'http://example.com/'))
+    >>> lx[0]
+    u'<a href="url1">link1</a>'
+    >>> lx[1]
+    u'<a href="url2">link2</a>'
+    >>> cl = SapeClient(user = 'user123456789', db_driver = ('mem',),
+    ...                 use_xml = True, xml_server_list = ['file:///does_not_exists'])
+    >>> cl.get_raw_links(PageRequest(url = 'http://example.com/'))
+    []
+    >>> cl.get_raw_links(PageRequest(url = 'http://example.com/'))
+    []
+    """
+
+    server_list = [
+            'http://dispenser-01.sape.ru/code.php?user=%(user)s&host=%(host)s&charset=utf-8',
+            'http://dispenser-02.sape.ru/code.php?user=%(user)s&host=%(host)s&charset=utf-8']
+    xml_server_list = [
+            'http://dispenser-01.sape.ru/code.php?user=%(user)s&host=%(host)s&charset=utf-8&as_xml=1',
+            'http://dispenser-02.sape.ru/code.php?user=%(user)s&host=%(host)s&charset=utf-8&as_xml=1']
+
+    def get_links_new_page(self, data, request):
+        if self.is_bot(data, request) or self.force_show_code:
+            try:
+                return [data['__sape_new_url__']]
+            except KeyError:
+                pass
+        return []
+
+    def get_delimiter(self, data, request):
+        return data.get('__sape_delimiter__', '')
+
+    def is_bot(self, data, request):
+        return request.cookies.get('sape_cookie') == self.user
+
+    def transform_code(self, data, request, code):
+        if self.is_bot(data, request):
+            code = '<sape_noindex>%s</sape_noindex>' % code
+        return code
+
+    def parse_data(self, data):
+        for key, value in data.items():
+            if key.startswith('/'):
+                if type(value) == dict:
+                    value = value.values()
+                yield (normalize_uri(key), map(self.parse_link, value))
+            else:
+                if type(value) == str:
+                    value = unicode(value, 'utf-8')
+                yield (key, value)
+
+    def parse_data_xml(self, data):
+        def node_text(node):
+            return u''.join([sn.nodeValue for sn in node.childNodes
+                if sn.nodeType == xml.dom.Node.TEXT_NODE])
+        path = []
+        for (event, node) in data:
+            if event == xml.dom.pulldom.START_ELEMENT:
+                path.append(node.tagName)
+                if path == ['sape', 'page']:
+                    data.expandNode(node)
+                    path.pop()
+                    uri = node.getAttribute('uri')
+                    if uri == '*':
+                        yield ('__sape_new_url__', node_text(node))
+                    else:
+                        uri = normalize_uri(uri)
+                        link_nodes = node.getElementsByTagName('link')
+                        yield (uri, [self.parse_link(node_text(x))
+                            for x in link_nodes])
+                elif path == ['sape', 'sape_ips']:
+                    data.expandNode(node)
+                    path.pop()
+                    ip_nodes = node.getElementsByTagName('ip')
+                    yield ('__sape_ips__',
+                            [node_text(x) for x in ip_nodes])
+                elif path == ['sape']:
+                    delimiter = node.getAttribute('delimiter')
+                    if delimiter:
+                        yield ('__sape_delimiter__', delimiter)
+            elif event == xml.dom.pulldom.END_ELEMENT:
+                path.pop()
+
 class ContextLinksGenerator(HTMLParser.HTMLParser):
     def __init__(self, out, links,
             is_fragment, show_code, new_url_code,
@@ -356,8 +543,8 @@ class ContextLinksGenerator(HTMLParser.HTMLParser):
         self.char_buf[:] = []
         if not self.exclude_ctx:
             if set(self.include_ctx) == self.include_tags:
-                for sentence_re, link_code in self.links:
-                    content = sentence_re.sub(link_code, content, count = 1)
+                for sentence_re, link in self.links:
+                    content = sentence_re.sub(link, content, count = 1)
         self.out.write(content)
 
     def handle_comment(self, data):
@@ -376,7 +563,8 @@ class SapeContextClient(SapeClient):
     >>> from linkexchange.clients.base import PageRequest
     >>> test_server = SapeTestServer()
     >>> cl = SapeContextClient(user = 'user123456789', db_driver = ('mem',),
-    ...                 server_list = [test_server.url], force_show_code = False)
+    ...                 use_xml = True, xml_server_list = [test_server.url],
+    ...                 force_show_code = False)
     >>> html = lambda x: u'<html>\\n<body>\\n%s\\n</body>\\n</html>' % x
     >>> req = PageRequest(url = 'http://example.com/',
     ...                   cookies = dict(sape_cookie = cl.user))
@@ -416,11 +604,15 @@ class SapeContextClient(SapeClient):
     </html>
     """
 
-    def __init__(self, user, db_driver, **kw):
-        kw.setdefault('force_show_code', False)
-        kw.setdefault('server_list', [
+    force_show_code = False
+    server_list = [
             'http://dispenser-01.sape.ru/code_context.php?user=%(user)s&host=%(host)s&charset=utf-8',
-            'http://dispenser-02.sape.ru/code_context.php?user=%(user)s&host=%(host)s&charset=utf-8'])
+            'http://dispenser-02.sape.ru/code_context.php?user=%(user)s&host=%(host)s&charset=utf-8']
+    xml_server_list = [
+            'http://dispenser-01.sape.ru/code_context.php?user=%(user)s&host=%(host)s&charset=utf-8&as_xml=1',
+            'http://dispenser-02.sape.ru/code_context.php?user=%(user)s&host=%(host)s&charset=utf-8&as_xml=1']
+
+    def __init__(self, user, db_driver, **kw):
         super(SapeContextClient, self).__init__(user, db_driver, **kw)
         self.tags_re = re.compile(r'<[^>]*>')
         self.start_doc_re = re.compile(r'^\s*<(\?xml|!DOCTYPE|html|HTML)\b', re.S)
@@ -432,9 +624,8 @@ class SapeContextClient(SapeClient):
         return u''
 
     def content_filter(self, request, content):
-        data = self._load_data(request)
-        show_code = (request.cookies.get('sape_cookie') == self.user or
-                self.force_show_code)
+        data = self.load_data(request)
+        show_code = self.is_bot(data, request) or self.force_show_code
         links = data.get(str(request.uri), [])
         force_body_sape_index = False
         include_tags = set()
@@ -459,43 +650,12 @@ class SapeContextClient(SapeClient):
         generator.close()
         return out.getvalue()
 
-    def _do_load_from_server(self, url):
-        def str2uni(value):
-            if type(value) == str:
-                value = unicode(value, 'utf-8')
-            return value
-        def compile(link_code):
-            link_code = str2uni(link_code)
-            sentence = re.escape(xml.sax.saxutils.unescape(
-                self.tags_re.sub('', link_code)))
-            sentence.replace(' ', r'(\s|(&nbsp;))+')
-            return (re.compile(sentence, re.S + re.UNICODE), link_code)
-        req = urllib2.Request(url)
-        req.add_header('User-Agent', self.user_agent)
-        try:
-            log.debug("Fetching: %s", url)
-            f = urlopen_with_timeout(req, self.socket_timeout)
-            raw_data = f.read()
-            f.close()
-        except (urllib2.URLError, httplib.HTTPException), e:
-            log.error("Network error: %s", str(e))
-            raise ClientNetworkError('Network error: %s' % str(e))
-        if raw_data.startswith('FATAL ERROR:'):
-            log.error("Server error: %s", raw_data)
-            raise ClientError(raw_data)
-        try:
-            data = phpserialize.loads(raw_data)
-        except ValueError, e:
-            log.error("Could not deserialize response from server: %s", str(e))
-            raise ClientDataError('Could not deserialize response '
-                    'from server: %s' % str(e))
-        for key, value in data.items():
-            if key.startswith('/'):
-                if type(value) == dict:
-                    value = value.values()
-                yield (key, map(compile, value))
-            else:
-                yield (key, str2uni(value))
+    def parse_link(self, link):
+        link = super(SapeContextClient, self).parse_link(link)
+        sentence = re.escape(xml.sax.saxutils.unescape(
+            self.tags_re.sub('', link)))
+        sentence.replace(' ', r'(\s|(&nbsp;))+')
+        return (re.compile(sentence, re.S + re.UNICODE), link)
 
 if __name__ == "__main__":
     import doctest
